# coding: utf8

import torch
import torch.nn as nn

from .modules import Flatten, PadMaxPool3d

"""
All the architectures are built here
"""


class Conv5_FC3(nn.Module):
    """
    Classifier for a binary classification task

    Image level architecture used on Minimal preprocessing
    """

    def __init__(self, dropout=0.5, n_classes=2):
        super(Conv5_FC3, self).__init__()
        # fmt: off
        self.features = nn.Sequential(
            nn.Conv3d(1, 8, 3, padding=1),
            nn.BatchNorm3d(8),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(8, 16, 3, padding=1),
            nn.BatchNorm3d(16),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(16, 32, 3, padding=1),
            nn.BatchNorm3d(32),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(32, 64, 3, padding=1),
            nn.BatchNorm3d(64),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(64, 128, 3, padding=1),
            nn.BatchNorm3d(128),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

        )

        self.classifier = nn.Sequential(
            Flatten(),
            nn.Dropout(p=dropout),

            nn.Linear(128 * 6 * 7 * 6, 1300),
            nn.ReLU(),

            nn.Linear(1300, 50),
            nn.ReLU(),

            nn.Linear(50, n_classes)

        )

        self.flattened_shape = [-1, 128, 6, 7, 6]
        # fmt: on

    def forward(self, x):
        x = self.features(x)
        x = self.classifier(x)

        return x


class VConv5_FC3(nn.Module):
    """
    Classifier for a binary classification task

    Image level architecture used on Minimal preprocessing
    """

    def __init__(self, dropout=0.5, n_classes=2):
        super(VConv5_FC3, self).__init__()
        # fmt: off
        self.features = nn.Sequential(
            nn.Conv3d(1, 8, 3, padding=1),
            nn.BatchNorm3d(8),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(8, 16, 3, padding=1),
            nn.BatchNorm3d(16),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(16, 32, 3, padding=1),
            nn.BatchNorm3d(32),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(32, 64, 3, padding=1),
            nn.BatchNorm3d(64),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(64, 128, 3, padding=1),
            nn.BatchNorm3d(128),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

        )

        self.fc_mu = nn.Sequential(
            Flatten(),
            nn.Linear(128 * 6 * 7 * 6, 1300)
        )

        self.fc_var = nn.Sequential(
            Flatten(),
            nn.Linear(128 * 6 * 7 * 6, 1300)
        )

        self.classifier = nn.Sequential(
            nn.Linear(1300, 50),
            nn.ReLU(),

            nn.Linear(50, n_classes)

        )

        self.flattened_shape = [-1, 128, 6, 7, 6]
        self.variational = True
        # fmt: on

    def forward(self, x):
        x = self.features(x)
        log_var = self.fc_var(x)
        mu = self.fc_mu(x)
        std = torch.exp(log_var / 2)
        if self.training:
            q = torch.distributions.Normal(mu, std)
            z = q.rsample()
        else:
            z = mu
        out = self.classifier(z)

        return z, mu, std, out


class Conv5_FC3_mni(nn.Module):
    """
    Classifier for a binary classification task

    Image level architecture used on Extensive preprocessing
    """

    def __init__(self, dropout=0.5, n_classes=2):
        super(Conv5_FC3_mni, self).__init__()
        # fmt: off
        self.features = nn.Sequential(
            nn.Conv3d(1, 8, 3, padding=1),
            nn.BatchNorm3d(8),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(8, 16, 3, padding=1),
            nn.BatchNorm3d(16),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(16, 32, 3, padding=1),
            nn.BatchNorm3d(32),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(32, 64, 3, padding=1),
            nn.BatchNorm3d(64),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(64, 128, 3, padding=1),
            nn.BatchNorm3d(128),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

        )

        self.classifier = nn.Sequential(
            Flatten(),
            nn.Dropout(p=dropout),

            nn.Linear(128 * 4 * 5 * 4, 1300),
            nn.ReLU(),

            nn.Linear(1300, 50),
            nn.ReLU(),

            nn.Linear(50, n_classes)

        )

        self.flattened_shape = [-1, 128, 4, 5, 4]
        # fmt: on

    def forward(self, x):
        x = self.features(x)
        x = self.classifier(x)

        return x


class Conv6_FC3(nn.Module):
    """
    Classifier for a binary classification task

    Image level architecture used on Minimal preprocessing
    """

    def __init__(self, dropout=0.5, n_classes=2):
        super(Conv6_FC3, self).__init__()
        # fmt: off
        self.features = nn.Sequential(
            nn.Conv3d(1, 8, 3, padding=1),
            nn.BatchNorm3d(8),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(8, 16, 3, padding=1),
            nn.BatchNorm3d(16),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(16, 32, 3, padding=1),
            nn.BatchNorm3d(32),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(32, 64, 3, padding=1),
            nn.BatchNorm3d(64),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(64, 128, 3, padding=1),
            nn.BatchNorm3d(128),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(128, 256, 3, padding=1),
            nn.BatchNorm3d(256),
            nn.ReLU(),
            PadMaxPool3d(2, 2),
        )

        self.classifier = nn.Sequential(
            Flatten(),
            nn.Dropout(p=dropout),

            nn.Linear(256 * 3 * 4 * 3, 1000),
            nn.ReLU(),

            nn.Linear(1000, 50),
            nn.ReLU(),

            nn.Linear(50, n_classes)

        )

        self.flattened_shape = [-1, 256, 3, 4, 3]
        # fmt: on

    def forward(self, x):
        x = self.features(x)
        x = self.classifier(x)

        return x


class Conv5_FC3_down(nn.Module):
    """
    Classifier for a binary classification task

    Image level architecture used on Minimal preprocessing
    """

    def __init__(self, dropout=0.5, n_classes=2):
        super(Conv5_FC3_down, self).__init__()
        # fmt: off
        self.features = nn.Sequential(
            nn.Conv3d(1, 8, 3, padding=1),
            nn.BatchNorm3d(8),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(8, 16, 3, padding=1),
            nn.BatchNorm3d(16),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(16, 32, 3, padding=1),
            nn.BatchNorm3d(32),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(32, 64, 3, padding=1),
            nn.BatchNorm3d(64),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

            nn.Conv3d(64, 128, 3, padding=1),
            nn.BatchNorm3d(128),
            nn.ReLU(),
            PadMaxPool3d(2, 2),

        )

        self.classifier = nn.Sequential(
            Flatten(),
            nn.Dropout(p=dropout),

            nn.Linear(128 * 3 * 4 * 3, 350),
            nn.ReLU(),

            nn.Linear(350, 25),
            nn.ReLU(),

            nn.Linear(25, n_classes)

        )

        self.flattened_shape = [-1, 128, 3, 4, 3]
        # fmt: on

    def forward(self, x):
        x = self.features(x)
        x = self.classifier(x)

        return x
